"""Definitions for the StatusUpdate model."""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _
from djblets.db.fields import JSONField

from reviewboard.changedescs.models import ChangeDescription
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.managers import StatusUpdateManager
from reviewboard.reviews.models.base_comment import BaseComment
from reviewboard.reviews.models.review import Review
from reviewboard.reviews.models.review_request import ReviewRequest
from reviewboard.reviews.signals import status_update_request_run
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from typing import ClassVar, Literal, Optional

    from django.contrib.auth.models import AnonymousUser
    from typelets.django.strings import StrOrPromise
    from typing_extensions import TypeAlias

    _StateFlag: TypeAlias = Literal['C', 'E', 'F', 'P', 'R', 'S', 'T']
    _StateStr: TypeAlias = Literal[
        'cancelled',
        'done-failure',
        'done-success',
        'error',
        'not-yet-run',
        'pending',
        'timed-out',
    ]


class StatusUpdate(models.Model):
    """A status update from a third-party service or extension.

    This status model allows external services (such as continuous integration
    services, Review Bot, etc.) to provide an update on their status. An
    example of this would be a CI tool which does experimental builds of
    changes. While the build is running, that tool would set its status to
    pending, and when it was done, would set it to one of the done states,
    and potentially associate it with a review containing issues.
    """

    #: The pending state.
    PENDING = 'P'

    #: The completed successfully state.
    DONE_SUCCESS = 'S'

    #: The completed with reported failures state.
    DONE_FAILURE = 'F'

    #: The error state.
    ERROR = 'E'

    #: Timeout state.
    TIMEOUT = 'T'

    #: Not yet run state.
    NOT_YET_RUN = 'R'

    #: Cancelled state.
    CANCELLED = 'C'

    STATUSES = (
        (PENDING, _('Pending')),
        (DONE_SUCCESS, _('Done (Success)')),
        (DONE_FAILURE, _('Done (Failure)')),
        (ERROR, _('Error')),
        (TIMEOUT, _('Timed Out')),
        (NOT_YET_RUN, _('Not Yet Run')),
        (CANCELLED, _('Cancelled')),
    )

    _INTEGRATION_CONFIG_KEY = '__integration_config_id'

    #: An identifier for the service posting this status update.
    #:
    #: This ID is self-assigned, and just needs to be unique to that service.
    #: Possible values can be an extension ID, webhook URL, or a script name.
    service_id = models.CharField(_('Service ID'), max_length=255)

    #: The user who created this status update.
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='status_updates',
        verbose_name=_('User'),
        blank=True,
        null=True)

    #: The timestamp of the status update.
    timestamp = models.DateTimeField(_('Timestamp'), auto_now=True)

    #: A user-visible short summary of the status update.
    #:
    #: This is typically the name of the integration or tool that was run.
    summary = models.CharField(_('Summary'), max_length=255)

    #: A user-visible description on the status update.
    #:
    #: This is shown in the UI adjacent to the summary. Typical results might
    #: be things like "running." or "failed.". This should include punctuation.
    description = models.CharField(_('Description'), max_length=255,
                                   blank=True)

    #: An optional link.
    #:
    #: This is used in case the tool has some external page, such as a build
    #: results page on a CI system.
    url = models.URLField(_('Link URL'), max_length=255, blank=True)

    #: Text for the link. If ``url`` is empty, this will not be used.
    url_text = models.CharField(_('Link text'), max_length=64, blank=True)

    #: The current state of this status update.
    #:
    #: This should be set to :py:attr:`PENDING` while the service is
    #: processing the update, and then to either :py:attr:`DONE_SUCCESS` or
    #: :py:attr:`DONE_FAILURE` once complete. If the service encountered some
    #: error which prevented completion, this should be set to
    #: :py:attr:`ERROR`.
    state = models.CharField(_('State'), max_length=1, choices=STATUSES)

    #: The review request that this status update is for.
    review_request = models.ForeignKey(
        ReviewRequest,
        on_delete=models.CASCADE,
        related_name='status_updates',
        verbose_name=_('Review Request'))

    #: The change to the review request that this status update is for.
    #:
    #: If this is ``None``, this status update refers to the review request as
    #: a whole (for example, the initial diff that was posted).
    change_description = models.ForeignKey(
        ChangeDescription,
        on_delete=models.CASCADE,
        related_name='status_updates',
        verbose_name=_('Change Description'),
        null=True,
        blank=True)

    #: An optional review created for this status update.
    #:
    #: This allows the third-party service to create comments and open issues.
    review = models.OneToOneField(
        Review,
        on_delete=models.CASCADE,
        related_name='status_update',
        verbose_name=_('Review'),
        null=True,
        blank=True)

    #: Any extra data that the service wants to store for this status update.
    extra_data = JSONField(null=True)

    #: An (optional) timeout, in seconds. If this is non-None and the state has
    #: been ``PENDING`` for longer than this period (computed from the
    #: :py:attr:`timestamp` field), :py:attr:`effective_state` will be
    #: ``TIMEOUT``.
    timeout = models.IntegerField(null=True, blank=True)

    objects: ClassVar[StatusUpdateManager] = StatusUpdateManager()

    ######################
    # Instance variables #
    ######################

    _integration_config: Optional[IntegrationConfig]

    @staticmethod
    def state_to_string(
        state: _StateFlag,
    ) -> _StateStr:
        """Return a string representation of a status update state.

        Args:
            state (str):
                A single-character string representing the state.

        Returns:
            str:
            A longer string representation of the state suitable for use in
            the API.
        """
        if state == StatusUpdate.PENDING:
            return 'pending'
        elif state == StatusUpdate.DONE_SUCCESS:
            return 'done-success'
        elif state == StatusUpdate.DONE_FAILURE:
            return 'done-failure'
        elif state == StatusUpdate.ERROR:
            return 'error'
        elif state == StatusUpdate.TIMEOUT:
            return 'timed-out'
        elif state == StatusUpdate.NOT_YET_RUN:
            return 'not-yet-run'
        elif state == StatusUpdate.CANCELLED:
            return 'cancelled'
        else:
            raise ValueError(f'Invalid state "{state}"')  # type:ignore

    @staticmethod
    def string_to_state(
        state: _StateStr,
    ) -> _StateFlag:
        """Return a status update state from an API string.

        Args:
            state (str):
                A string from the API representing the state.

        Returns:
            str:
            A single-character string representing the state, suitable for
            storage in the ``state`` field.
        """
        if state == 'pending':
            return StatusUpdate.PENDING
        elif state == 'done-success':
            return StatusUpdate.DONE_SUCCESS
        elif state == 'done-failure':
            return StatusUpdate.DONE_FAILURE
        elif state == 'error':
            return StatusUpdate.ERROR
        elif state == 'timed-out':
            return StatusUpdate.TIMEOUT
        elif state == 'not-yet-run':
            return StatusUpdate.NOT_YET_RUN
        elif state == 'cancelled':
            return StatusUpdate.CANCELLED
        else:
            raise ValueError(f'Invalid state string "{state}"')  # type:ignore

    @property
    def integration_config(self) -> Optional[IntegrationConfig]:
        """The integration config that manages this status update, if any.

        If the stored configuration no longer exists, or is no longer
        applicable to any associated :term:`Local Site`, then this will be
        ``None``.

        The configuration is cached for repeated lookups.

        Version Added:
            5.0.3

        Type:
            reviewboard.integrations.models.IntegrationConfig
        """
        config: Optional[IntegrationConfig]

        if hasattr(self, '_integration_config'):
            config = self._integration_config
        else:
            config = None
            config_id = self.extra_data.get(self._INTEGRATION_CONFIG_KEY)

            if config_id:
                try:
                    config = (
                        IntegrationConfig.objects
                        .get(Q(pk=config_id) &
                             LocalSite.objects.build_q(
                                 self.review_request.local_site_id,
                                 allow_all=False))
                    )
                except IntegrationConfig.DoesNotExist:
                    # Either the configuration was removed, or it's on the
                    # wrong Local Site. Fall back to None.
                    pass

            self._integration_config = config

        return config

    @integration_config.setter
    def integration_config(
        self,
        config: IntegrationConfig | None,
    ) -> None:
        """Set the integration configuration for this status update.

        This allows integrations to properly use the correct configuration
        when manually running for the status update.

        Version Added:
            5.0.3

        Args:
            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration to store.

                The configuration's :term:`Local Site` must match that of
                the status update.

        Raises:
            ValueError:
                The provided configuration value is not supported by this
                status update.
        """
        if config:
            if config.local_site_id != self.review_request.local_site_id:
                raise ValueError(
                    'The integration configuration and Status Update must '
                    'have the same Local Site.')

            self.extra_data[self._INTEGRATION_CONFIG_KEY] = config.pk
        else:
            self.extra_data.pop(self._INTEGRATION_CONFIG_KEY, None)

        self._integration_config = config

    def is_mutable_by(
        self,
        user: User | AnonymousUser,
    ) -> bool:
        """Return whether the user can modify this status update.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            True if the user can modify this status update.
        """
        user_pk = user.pk
        review_request = self.review_request

        return (user.is_authenticated and
                (self.user_id == user_pk or
                 user_pk == review_request.submitter_id or
                 user.has_perm('reviews.change_statusupdate',
                               review_request.local_site)))

    @property
    def effective_state(self) -> _StateFlag:
        """The state of the status update, taking into account timeouts.

        Type:
            str
        """
        if self.state == self.PENDING and self.timeout is not None:
            timeout = self.timestamp + datetime.timedelta(seconds=self.timeout)

            if timezone.now() > timeout:
                return self.TIMEOUT

        return self.state

    def drop_open_issues(self) -> None:
        """Drop any open issues associated with this status update."""
        if self.review is None:
            return

        now = timezone.now()
        review_updated = False

        for comments in (self.review.comments,
                         self.review.screenshot_comments,
                         self.review.file_attachment_comments,
                         self.review.general_comments):
            open_comments = comments.filter(issue_status=BaseComment.OPEN)
            count = open_comments.update(issue_status=BaseComment.DROPPED,
                                         timestamp=now)

            if count > 0:
                review_updated = True

        if review_updated:
            self.review_request.last_review_activity_timestamp = now
            self.review_request.save(
                update_fields=['last_review_activity_timestamp'])
            self.review_request.reinit_issue_open_count()

    @property
    def can_run(self) -> bool:
        """Whether or not the checker associated can be run.

        Type:
            bool
        """
        state = self.effective_state
        return (state == StatusUpdate.NOT_YET_RUN or
                (state in {StatusUpdate.ERROR, StatusUpdate.TIMEOUT} and
                 self.extra_data.get('can_retry')))

    @property
    def action_name(self) -> StrOrPromise:
        """The name of the action to use for running or re-running the check.

        Type:
            str
        """
        if self.effective_state in {StatusUpdate.ERROR, StatusUpdate.TIMEOUT}:
            return gettext('Retry')
        else:
            return gettext('Run')

    def run(self) -> None:
        """Run the tool associated with this status update.

        This will emit the :py:data:`~reviewboard.reviews.signals.
        status_update_request_run` signal, which extensions/integrations
        providing manual run support should listen to. They're responsible
        for handling any filtering and configuration matching, as required.

        Version Changed:
            5.0.3:
            An associated integration config (if stored along with the status
            update) will be provided to the signal.
        """
        assert self.can_run
        status_update_request_run.send(sender=self.__class__,
                                       status_update=self,
                                       config=self.integration_config)

    class Meta:
        """Metadata for the model."""

        app_label = 'reviews'
        db_table = 'reviews_statusupdate'
        ordering = ['timestamp']
        get_latest_by = 'timestamp'
        verbose_name = _('Status Update')
        verbose_name_plural = _('Status Updates')



from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.webapi.models import WebAPIToken


class WebAPITokenAdmin(ModelAdmin):
    list_display = ('user', 'local_site', 'time_added', 'last_updated', 'note')
    raw_id_fields = ('user', 'local_site')


admin_site.register(WebAPIToken, WebAPITokenAdmin)



"""Unit tests for RootDiffCommentResource."""

from datetime import timedelta
from django.utils import timezone
from djblets.testing.decorators import add_fixtures
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import \
    review_diff_comment_list_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import get_root_diff_comment_list_url


class ResourceListTests(BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the RootDiffCommentResource API."""

    fixtures = ['test_scmtools', 'test_users']
    sample_api_url = 'diff-comments/'
    resource = resources.root_diff_comment
    test_http_methods = ('GET',)

    def compare_item(self, item_rsp, comment):
        """Compare a comment item to a comment returned from the API.

        Args:
            item_rsp (dict):
                The serialized comment returned from the API.

            comment (reviewboard.reviews.models.Comment):
                The comment instance to compare to.

        Raises:
            AssertionError:
                The API response was not equivalent to the object.
        """
        self.assertEqual(item_rsp['id'], comment.pk)
        self.assertEqual(item_rsp['text'], comment.text)
        self.assertEqual(item_rsp['issue_opened'], comment.issue_opened)
        self.assertEqual(item_rsp['first_line'], comment.first_line)
        self.assertEqual(item_rsp['num_lines'], comment.num_lines)

        self.assertEqual(item_rsp['extra_data'],
                         self.resource._strip_private_data(comment.extra_data))

        if comment.rich_text:
            self.assertEqual(item_rsp['text_type'], 'markdown')
        else:
            self.assertEqual(item_rsp['text_type'], 'plain')

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)

        if populate_items:
            items = [self.create_diff_comment(review, filediff)]
        else:
            items = []

        return (get_root_diff_comment_list_url(local_site_name),
                review_diff_comment_list_mimetype,
                items)

    @webapi_test_template
    def test_comment_is_same_as_canonical_comment(self):
        """Testing the GET <URL>/ API returns a proper data type"""
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        review = self.create_review(review_request, publish=True)

        comment = self._create_diff_comment(review_request, review)

        rsp = self.api_get(get_root_diff_comment_list_url(), {},
                           expected_mimetype=review_diff_comment_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp[self.resource.list_result_key][0], comment)

    @webapi_test_template
    def test_get_with_counts_only(self):
        """"Testing the GET <URL>/?counts-only=1 API"""
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        review = self.create_review(review_request, publish=True)

        self._create_diff_comment(review_request, review)

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'counts-only': 1,
        }, expected_mimetype=review_diff_comment_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['count'], review.comments.count())

    @webapi_test_template
    def test_get(self):
        """"Testing the GET <URL>/? API returns only comments from reviews
        and review requests that the requester has access to
        """
        # Publicly-accessible published review request.
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)

        # Comment from a published review on a publicly-accessible
        # review request.
        review1 = self.create_review(review_request, publish=True)
        comment1 = self._create_diff_comment(review_request, review1)

        # Comment from an unpublished review on a publicly-accessible
        # review request.
        review2 = self.create_review(review_request, publish=False)
        self._create_diff_comment(review_request, review2)

        # Comment from a published review owned by the requester on a
        # publicly-accessible review request.
        review3 = self.create_review(review_request,
                                     user=self.user,
                                     publish=True)
        comment3 = self._create_diff_comment(review_request, review3)

        # Comment from an unpublished review owned by the requester on a
        # publicly-accessible review request.
        review4 = self.create_review(review_request,
                                     user=self.user,
                                     publish=False)
        comment4 = self._create_diff_comment(review_request, review4)

        # Published review request from a private repository the requester
        # does not have access to.
        repo = self.create_repository(public=False)
        review_request_inaccessible = self.create_review_request(
            repository=repo,
            publish=True)

        # Comment from a published review on a private repository the requester
        # does not have access to.
        review5 = self.create_review(review_request_inaccessible, publish=True)
        self._create_diff_comment(review_request_inaccessible, review5)

        # Comment from an unpublished review on a private repository the
        # requester does not have access to.
        review6 = self.create_review(review_request_inaccessible,
                                     publish=False)
        self._create_diff_comment(review_request_inaccessible, review6)

        # An invite-only review group used to limit access for the following
        # review requests.
        group = self.create_review_group(invite_only=True)

        # Comment from a published review from a review request that has an
        # invite-only review group not accessible to the requester, but they
        # have access to through being a targeted reviewer.
        review_request_targeted = self.create_review_request(
            publish=True,
            create_repository=True)
        review_request_targeted.target_groups.add(group)
        review_request_targeted.target_people.add(self.user)
        review7 = self.create_review(review_request_targeted, publish=True)
        comment7 = self._create_diff_comment(review_request_targeted,
                                             review7)

        # Comment from a published review from a review request that has an
        # invite-only review group not accessible to the requester, and that
        # they do not have access to because they are not listed as a
        # target reviewer.
        review_request_untargeted = self.create_review_request(
            publish=True,
            create_repository=True)
        review_request_untargeted.target_groups.add(group)
        review8 = self.create_review(review_request_untargeted, publish=True)
        self._create_diff_comment(review_request_untargeted, review8)

        rsp = self.api_get(get_root_diff_comment_list_url(), {},
                           expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 4)
        self.compare_item(rsp_items[0], comment1)
        self.compare_item(rsp_items[1], comment3)
        self.compare_item(rsp_items[2], comment4)
        self.compare_item(rsp_items[3], comment7)

    @webapi_test_template
    def test_get_by_repo(self):
        """Testing the GET <URL>/ API returns only comments from
        repositories that the requester has access to
        """
        # Comment from a public repository.
        repo1 = self.create_repository(name='repo1', public=True)
        review_request1 = self.create_review_request(publish=True,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self._create_diff_comment(review_request1, review1)

        # Comment from a private repository that the requester has
        # access to from being listed in the repository's users list.
        repo2 = self.create_repository(name='repo2', public=False)
        repo2.users.add(self.user)
        review_request2 = self.create_review_request(publish=True,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)
        comment2 = self._create_diff_comment(review_request2, review2)

        # An invite-only review group that the requester has access to.
        group_accessible = self.create_review_group(invite_only=True)
        group_accessible.users.add(self.user)

        # Comment from a private repository that the requester has
        # access to through being a member of a targeted review group.
        repo3 = self.create_repository(name='repo3', public=False)
        repo3.review_groups.add(group_accessible)
        review_request3 = self.create_review_request(publish=True,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)
        comment3 = self._create_diff_comment(review_request3, review3)

        # Comment from a private repository that the requester does
        # not have access to.
        repo4 = self.create_repository(name='repo4', public=False)
        review_request4 = self.create_review_request(publish=True,
                                                     repository=repo4)
        review4 = self.create_review(review_request4, publish=True)
        self._create_diff_comment(review_request4, review4)

        # Comment from a private repository that the requester has access
        # to through being a member of a targeted review group and
        # being listed on the repository's users list.
        repo5 = self.create_repository(name='repo5', public=False)
        repo5.review_groups.add(group_accessible)
        repo5.users.add(self.user)
        review_request5 = self.create_review_request(publish=True,
                                                     repository=repo5)
        review5 = self.create_review(review_request5, publish=True)
        comment5 = self._create_diff_comment(review_request5, review5)

        # An invite-only review group that the requester does not have
        # access to.
        group_inaccessible = self.create_review_group(invite_only=True)

        # Comment from a private repository that targets an invite-only review
        # group, but that the requester has access to from being listed in the
        # repository's users list.
        repo6 = self.create_repository(name='repo6', public=False)
        repo6.review_groups.add(group_inaccessible)
        repo6.users.add(self.user)
        review_request6 = self.create_review_request(publish=True,
                                                     repository=repo6)
        review6 = self.create_review(review_request6, publish=True)
        comment6 = self._create_diff_comment(review_request6, review6)

        # Comment from a private repository that targets an invite-only review
        # group and that the requester does not have access to.
        repo7 = self.create_repository(name='repo7', public=False)
        repo7.review_groups.add(group_inaccessible)
        review_request7 = self.create_review_request(publish=True,
                                                     repository=repo7)
        review7 = self.create_review(review_request7, publish=True)
        self._create_diff_comment(review_request7, review7)

        rsp = self.api_get(get_root_diff_comment_list_url(), {},
                           expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 5)
        self.compare_item(rsp_items[0], comment1)
        self.compare_item(rsp_items[1], comment2)
        self.compare_item(rsp_items[2], comment3)
        self.compare_item(rsp_items[3], comment5)
        self.compare_item(rsp_items[4], comment6)

    @webapi_test_template
    def test_get_by_review_group(self):
        """Testing the GET <URL>/ API returns only comments associated
        with review groups that the requester has access to
        """
        # Comment that the requester has access to from being in a public
        # review group that is targeted by the review request.
        group1 = self.create_review_group(name='group1', invite_only=False)
        group1.users.add(self.user)
        review_request1 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request1.target_groups.add(group1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self._create_diff_comment(review_request1, review1)

        # Comment that the requester has access to from being in an invite-only
        # review group that is targeted by the review request.
        group2 = self.create_review_group(name='group2', invite_only=True)
        group2.users.add(self.user)
        review_request2 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request2.target_groups.add(group2)
        review2 = self.create_review(review_request2, publish=True)
        comment2 = self._create_diff_comment(review_request2, review2)

        # Comment that the requester has access to since there is a public
        # review group that is targeted by the review request.
        group3 = self.create_review_group(name='group3', invite_only=False)
        review_request3 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request3.target_groups.add(group3)
        review3 = self.create_review(review_request3, publish=True)
        comment3 = self._create_diff_comment(review_request3, review3)

        # Comment that the requester does not have access to since there is an
        # invite-only review group that is targeted by the review request.
        group4 = self.create_review_group(name='group4', invite_only=True)
        review_request4 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request4.target_groups.add(group4)
        review4 = self.create_review(review_request4, publish=True)
        self._create_diff_comment(review_request4, review4)

        rsp = self.api_get(get_root_diff_comment_list_url(), {},
                           expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 3)
        self.compare_item(rsp_items[0], comment1)
        self.compare_item(rsp_items[1], comment2)
        self.compare_item(rsp_items[2], comment3)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_by_local_site(self):
        """Testing the GET <URL> API returns only comments from local
        sites that the requester has access to
        """
        # Comment from a private local site that the requester has access to.
        local_site1 = self.get_local_site(self.local_site_name)
        local_site1.users.add(self.user)
        repo1 = self.create_repository(local_site=local_site1)
        review_request1 = self.create_review_request(publish=True,
                                                     local_site=local_site1,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self._create_diff_comment(review_request1, review1)

        # Comment from a private local site that the requester does not have
        # access to.
        local_site2 = self.get_local_site('local-site-2')
        repo2 = self.create_repository(local_site=local_site2)
        review_request2 = self.create_review_request(publish=True,
                                                     local_site=local_site2,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)
        self._create_diff_comment(review_request2, review2)

        # Comment from a public local site.
        local_site3 = self.create_local_site('public-local-site', public=True)
        repo3 = self.create_repository(local_site=local_site3)
        review_request3 = self.create_review_request(publish=True,
                                                     local_site=local_site3,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)
        self._create_diff_comment(review_request3, review3)

        # Comment from a global site.
        review_request4 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review4 = self.create_review(review_request4, publish=True)
        comment4 = self._create_diff_comment(review_request4, review4)

        # Testing that only comments from the given local site are returned.
        rsp = self.api_get(
            get_root_diff_comment_list_url(local_site_name=local_site1.name),
            {},
            expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

        # Testing that the requester is not authorized to make requests
        # to a local site that they do not have access to.
        rsp = self.api_get(
            get_root_diff_comment_list_url(local_site_name=local_site2.name),
            {'counts-only': 1},
            expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')

        # Testing that comments from local sites do not leak into results
        # from the global site.
        rsp = self.api_get(get_root_diff_comment_list_url(), {},
                           expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment4)

    @webapi_test_template
    def test_get_with_review_id(self):
        """"Testing the GET <URL>/?review-id=<id> API returns only comments
        from reviews that the requester has access to
        """
        # Publicly-accessible published review request.
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)

        # Comment from a published review on a publicly-accessible
        # review request.
        review1 = self.create_review(review_request, publish=True)
        comment1 = self._create_diff_comment(review_request, review1)

        # Comment from an unpublished review on a publicly-accessible
        # review request.
        review2 = self.create_review(review_request, publish=False)
        self._create_diff_comment(review_request, review2)

        # Comment from a published review owned by the requester on a
        # publicly-accessible review request.
        review3 = self.create_review(review_request,
                                     user=self.user,
                                     publish=True)
        self._create_diff_comment(review_request, review3)

        # Comment from an unpublished review owned by the requester on a
        # publicly-accessible review request.
        review4 = self.create_review(review_request,
                                     user=self.user,
                                     publish=False)
        comment4 = self._create_diff_comment(review_request, review4)

        # Published review request from a private repository the requester
        # does not have access to.
        repo = self.create_repository(public=False)
        review_request_inaccessible = self.create_review_request(
            repository=repo,
            publish=True)

        # Comment from a published review on a private repository the requester
        # does not have access to.
        review5 = self.create_review(review_request_inaccessible, publish=True)
        self._create_diff_comment(review_request_inaccessible, review5)

        # Comment from an unpublished review on a private repository the
        # requester does not have access to.
        review6 = self.create_review(review_request_inaccessible,
                                     publish=False)
        self._create_diff_comment(review_request_inaccessible, review6)

        # Testing that only comments from the given review are returned.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-id': review1.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

        # Testing that no comments get returned when querying for an
        # unpublished review that the requester does not have access to.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-id': review2.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

        # Testing that comments from an unpublished review that the requester
        # has access to get returned.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-id': review4.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment4)

        # Testing that no comments get returned when querying for a published
        # review on a private repository the requester does not have access to.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-id': review5.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

        # Testing that no comments get returned when querying for a review
        # that doesn't exist.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-id': -1,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @webapi_test_template
    def test_get_with_review_request(self):
        """"Testing the GET <URL>/?review-request-id=<id> API returns
        only comments from review requests that the requester has access to
        """
        # Publicly-accessible published review request.
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)

        # Comment from a published review on a publicly-accessible
        # review request.
        review1 = self.create_review(review_request, publish=True)
        comment1 = self._create_diff_comment(review_request, review1)

        # Comment from an unpublished review on a publicly-accessible
        # review request.
        review2 = self.create_review(review_request, publish=False)
        self._create_diff_comment(review_request, review2)

        # Comment from a published review owned by the requester on a
        # publicly-accessible review request.
        review3 = self.create_review(review_request,
                                     user=self.user,
                                     publish=True)
        comment3 = self._create_diff_comment(review_request, review3)

        # Comment from an unpublished review owned by the requester on a
        # publicly-accessible review request.
        review4 = self.create_review(review_request,
                                     user=self.user,
                                     publish=False)
        comment4 = self._create_diff_comment(review_request, review4)

        # Published review request from a private repository the requester
        # does not have access to.
        repo = self.create_repository(public=False)
        review_request_inaccessible = self.create_review_request(
            repository=repo,
            publish=True)

        # Comment from a published review on a private repository the requester
        # does not have access to.
        review5 = self.create_review(review_request_inaccessible, publish=True)
        self._create_diff_comment(review_request_inaccessible, review5)

        # Comment from an unpublished review on a private repository the
        # requester does not have access to.
        review6 = self.create_review(review_request_inaccessible,
                                     publish=False)
        self._create_diff_comment(review_request_inaccessible, review6)

        # Testing that only comments from the given review request
        # are returned.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-request-id': review_request.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 3)
        self.compare_item(rsp_items[0], comment1)
        self.compare_item(rsp_items[1], comment3)
        self.compare_item(rsp_items[2], comment4)

        # Testing that no comments are returned when the requester does
        # not have access to the given review request.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-request-id': review_request_inaccessible.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_with_review_request_from_local_site(self):
        """"Testing the GET <URL>/?review-request-id=<id> API with a LocalSite
        bound review request ID
        """
        # Comment from a review request on the global site.
        review_request1 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self._create_diff_comment(review_request1, review1)

        # Comment from a review request on a local site.
        local_site1 = self.get_local_site(self.local_site_name)
        local_site1.users.add(self.user)
        review_request2 = self.create_review_request(publish=True,
                                                     local_site=local_site1,
                                                     create_repository=True)
        review2 = self.create_review(review_request2, user=self.user,
                                     publish=True)
        comment2 = self._create_diff_comment(review_request2, review2,
                                             text='localsite')

        # Testing that passing the local ID of a review request from a local
        # site properly returns comments from that review request.
        rsp = self.api_get(
            get_root_diff_comment_list_url(local_site_name=local_site1.name),
            {'review-request-id': review_request2.local_id},
            expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment2)

        # Testing that comments from the local site review request do not leak
        # into the results when querying for a review request whose global ID
        # is the same as the local ID of the local site review request.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'review-request-id': review_request1.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

    @webapi_test_template
    def test_get_with_user(self):
        """"Testing the GET <URL>/?user=<user-name> API returns comments
        by user
        """
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)

        user1 = self.create_user(username='user1')

        review1 = self.create_review(review_request, publish=True, user=user1)
        comment1 = self._create_diff_comment(review_request, review1)

        review2 = self.create_review(review_request, publish=False, user=user1)
        self._create_diff_comment(review_request, review2)

        user2 = self.create_user(username='user2')

        review3 = self.create_review(review_request, publish=True, user=user2)
        self._create_diff_comment(review_request, review3)

        review4 = self.create_review(review_request, publish=False, user=user2)
        self._create_diff_comment(review_request, review4)

        # Testing that only comments by the given user that the requester
        # has access to are returned.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'user': 'user1',
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

        # Testing that no comments are returned when the given user
        # doesn't exist.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'user': 'user3',
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @webapi_test_template
    def test_get_with_file_diff_id(self):
        """"Testing the GET <URL>/?file-diff-id=<id> API"""
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        review = self.create_review(review_request, publish=True)
        diffset1 = self.create_diffset(review_request)

        # A comment.
        filediff1 = self.create_filediff(diffset1)
        comment1 = self.create_diff_comment(review, filediff1)

        # A comment linked to the same diffset but different file diff.
        filediff2 = self.create_filediff(diffset1)
        self.create_diff_comment(review, filediff2)

        # A comment from a different diffset and file diff.
        diffset2 = self.create_diffset(review_request, revision=2)
        filediff3 = self.create_filediff(diffset2)
        self.create_diff_comment(review, filediff3)

        # Testing that only comments from the given file diff are returned.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'file-diff-id': filediff1.id,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

        # Testing that no comments are returned when the given file diff
        # does not exist.
        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'file-diff-id': -1,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 0)

    @webapi_test_template
    def test_get_by_last_updated_from_includes_from_date(self):
        """Testing the GET <URL>/?last-updated-from=<date> API
        returns only comments within the from date
        """
        now = timezone.now()

        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset=diffset)
        review = self.create_review(review_request, publish=True,
                                    timestamp=(now - timedelta(hours=1)))

        # Comment created within the from date.
        comment1 = self.create_diff_comment(
            review,
            filediff,
            timestamp=(now + timedelta(hours=1)))

        # Comment created on the from date.
        comment2 = self.create_diff_comment(
            review,
            filediff,
            timestamp=now)

        # Comment created outside of the from date.
        self.create_diff_comment(
            review,
            filediff,
            timestamp=(now - timedelta(hours=1)))

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'last-updated-from': now.isoformat(),
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 2)
        self.compare_item(rsp_items[0], comment2)
        self.compare_item(rsp_items[1], comment1)

    @webapi_test_template
    def test_get_by_last_updated_to_includes_to_date(self):
        """Testing the GET <URL>/?last-updated-to=<date>
        API returns only comments within the to date
        """
        now = timezone.now()

        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset=diffset)
        review = self.create_review(review_request, publish=True,
                                    timestamp=(now - timedelta(hours=1)))

        # Comment created within the to date.
        comment1 = self.create_diff_comment(
            review,
            filediff,
            timestamp=(now - timedelta(hours=1)))

        # Comment created on the to date.
        self.create_diff_comment(review, filediff, timestamp=now)

        # Comment created outside of the to date.
        self.create_diff_comment(
            review,
            filediff,
            timestamp=(now + timedelta(hours=1)))

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'last-updated-to': now.isoformat(),
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment1)

    @webapi_test_template
    def test_get_by_last_updated_to_and_from_simple_date_format(self):
        """Testing the GET
        <URL>/?last-updated-from=<date>&last-updated-to=<date>
        API returns only comments within the from and to date
        """
        timestamp_from = timezone.now()
        timestamp_to = timezone.now() + timedelta(hours=1)

        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset=diffset)
        review = self.create_review(
            review_request,
            publish=True,
            timestamp=(timestamp_from - timedelta(minutes=30)))

        # Comment created within the from and to date.
        comment = self.create_diff_comment(
            review,
            filediff,
            timestamp=(timestamp_from + timedelta(minutes=30)))

        # Comment created before the from date.
        self.create_diff_comment(
            review,
            filediff,
            timestamp=(timestamp_from - timedelta(minutes=30)))

        # Comment created after the to date.
        self.create_diff_comment(
            review,
            filediff,
            timestamp=(timestamp_to + timedelta(minutes=30)))

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'last-updated-from': timestamp_from.isoformat(),
            'last-updated-to': timestamp_to.isoformat(),
        }, expected_mimetype=review_diff_comment_list_mimetype)
        rsp_items = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(rsp_items[0], comment)

    @webapi_test_template
    def test_get_with_is_reply(self):
        """Testing the GET <URL>/?is-reply=1 API"""
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        review = self.create_review(review_request, publish=True)

        comment1 = self._create_diff_comment(review_request, review)

        comment2 = self._create_diff_comment(review_request, review,
                                             reply_to=comment1)

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'is-reply': 1,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        items_rsp = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(items_rsp[0], comment2)

    @webapi_test_template
    def test_get_with_interdiff(self):
        """Testing the GET <URL>/?interdiff-revision= API"""
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        review = self.create_review(review_request, publish=True)

        # A first diff set and file diff.
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        # An interdiff and the second diff.
        interdiffset = self.create_diffset(review_request)
        interfilediff = self.create_filediff(diffset)

        # A comment linked to the interdiff.
        comment = self.create_diff_comment(review, filediff, interfilediff)

        # A comment not linked to the interdiff.
        self.create_diff_comment(review, filediff)

        rsp = self.api_get(get_root_diff_comment_list_url(), {
            'interdiff-revision': interdiffset.revision,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        items_rsp = rsp[self.resource.list_result_key]

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['total_results'], 1)
        self.compare_item(items_rsp[0], comment)

    def _create_diff_comment(self, review_request, review, **kwargs):
        """Creates a diff comment.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that the comment will be on.

            review (reviewboard.reviews.models.Review):
                The review that the comment will be on.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`create_diff_comment`.

        Returns:
            reviewboard.reviews.models.Comment:
            The comment.
        """
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset=diffset)

        return self.create_diff_comment(review, filediff, **kwargs)



import os
import unittest

from djblets.testing.decorators import add_fixtures
from djblets.util.filesystem import is_exe_in_path

from reviewboard.scmtools.bzr import BZRTool, get_bzr_exe
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         InvalidRevisionFormatError,
                                         RepositoryNotFoundError, SCMError)
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.testing.testcase import TestCase


bzr_exe = get_bzr_exe()


class BZRTests(SCMTestCase):
    """Unit tests for bzr."""

    fixtures = ['test_scmtools']

    ssh_required_system_exes = [bzr_exe]

    @unittest.skipIf(not is_exe_in_path(bzr_exe),
                     'Neither brz (Breezy) nor bzr (Bazaar) are installed')
    def setUp(self):
        super(BZRTests, self).setUp()

        self.bzr_repo_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), '..', 'testdata', 'bzr_repo'))
        self.bzr_ssh_path = ('bzr+ssh://localhost/%s'
                             % self.bzr_repo_path.replace('\\', '/'))
        self.bzr_sftp_path = ('sftp://localhost/%s'
                              % self.bzr_repo_path.replace('\\', '/'))
        self.repository = self.create_repository(
            name='Bazaar',
            path='file://' + self.bzr_repo_path,
            tool_name='Bazaar')

        self.tool = self.repository.get_scmtool()

    def test_check_repository(self):
        """Testing BZRTool.check_repository"""
        self.tool.check_repository(self.repository.path)

    def test_check_repository_with_not_found(self):
        """Testing BZRTool.check_repository with repository not found"""
        with self.assertRaises(RepositoryNotFoundError):
            self.tool.check_repository('file:///dummy')

    def test_ssh(self):
        """Testing BZRTool with a SSH-backed repository"""
        self._test_ssh(self.bzr_ssh_path, 'README')

    def test_ssh_with_site(self):
        """Testing BZRTool with a SSH-backed repository with a LocalSite"""
        self._test_ssh_with_site(self.bzr_ssh_path, 'README')

    def test_sftp(self):
        """Testing BZRTool with a SFTP-backed repository"""
        try:
            self._test_ssh(self.bzr_sftp_path, 'README')
        except SCMError as e:
            err = str(e)

            if 'Installed bzr and paramiko modules are incompatible' in err:
                raise unittest.SkipTest(err)

            raise

    def test_get_file(self):
        """Testing BZRTool.get_file"""
        content = self.tool.get_file('README', '2011-02-02 10:53:04 +0000')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_timezone_offset(self):
        """Testing BZRTool.get_file with timezone offset"""
        content = self.tool.get_file('README', '2011-02-02 02:53:04 -0800')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_non_utc_server_timezone(self):
        """Testing BZRTool.get_file with settings.TIME_ZONE != UTC"""
        old_timezone = os.environ[str('TZ')]
        os.environ[str('TZ')] = str('US/Pacific')

        try:
            content = self.tool.get_file('README', '2011-02-02 02:53:04 -0800')
        finally:
            os.environ[str('TZ')] = old_timezone

        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_revision_id(self):
        """Testing BZRTool.get_file with revision ID"""
        content = self.tool.get_file(
            'README',
            'revid:chipx86@chipx86.com-20110202105304-8lkgyb18aqr11b21')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_unknown_file(self):
        """Testing BZRTool.get_file with unknown file"""
        with self.assertRaises(FileNotFoundError):
            self.tool.get_file('NOT_REAL', '2011-02-02 02:53:04 -0800')

    def test_get_file_with_unknown_revision(self):
        """Testing BZRTool.get_file with unknown revision"""
        with self.assertRaises(FileNotFoundError):
            self.tool.get_file('README', '9999-02-02 02:53:04 -0800')

    def test_get_file_with_invalid_revision(self):
        """Testing BZRTool.get_file with invalid revision"""
        with self.assertRaises(InvalidRevisionFormatError):
            self.tool.get_file('README', r'\o/')

    def test_file_exists(self):
        """Testing BZRTool.files_exists"""
        self.assertTrue(self.tool.file_exists(
            'README',
            '2011-02-02 10:53:04 +0000'))

        self.assertFalse(self.tool.file_exists(
            'NOT_REAL',
            '2011-02-02 10:53:04 +0000'))
        self.assertFalse(self.tool.file_exists(
            'README',
            '9999-02-02 10:53:04 +0000'))

    def test_file_exists_with_timezone_offset(self):
        """Testing BZRTool.files_exists with timezone offset"""
        self.assertTrue(self.tool.file_exists(
            'README',
            '2011-02-02 02:53:04 -0800'))

    def test_file_exists_with_non_utc_server_timezone(self):
        """Testing BZRTool.files_exists with settings.TIME_ZONE != UTC"""
        old_timezone = os.environ[str('TZ')]
        os.environ[str('TZ')] = str('US/Pacific')

        try:
            self.assertTrue(self.tool.file_exists(
                'README',
                '2011-02-02 02:53:04 -0800'))
        finally:
            os.environ[str('TZ')] = old_timezone

    def test_file_exists_with_revision_id(self):
        """Testing BZRTool.files_exists with revision ID"""
        self.assertTrue(self.tool.file_exists(
            'README',
            'revid:chipx86@chipx86.com-20110202105304-8lkgyb18aqr11b21'))

    def test_file_exists_with_invalid_revision(self):
        """Testing BZRTool.files_exists with invalid revision"""
        with self.assertRaises(InvalidRevisionFormatError):
            self.tool.file_exists('README', r'\o/')


class BZRAuthFormTests(TestCase):
    """Unit tests for BZRTool's authentication form."""

    def test_fields(self):
        """Testing BZRTool authentication form fields"""
        form = BZRTool.create_auth_form()

        self.assertEqual(list(form.fields), ['username', 'password'])
        self.assertEqual(form['username'].help_text, '')
        self.assertEqual(form['username'].label, 'Username')
        self.assertEqual(form['password'].help_text, '')
        self.assertEqual(form['password'].label, 'Password')

    @add_fixtures(['test_scmtools'])
    def test_load(self):
        """Tetting BZRTool authentication form load"""
        repository = self.create_repository(
            tool_name='Bazaar',
            username='test-user',
            password='test-pass')

        form = BZRTool.create_auth_form(repository=repository)
        form.load()

        self.assertEqual(form['username'].value(), 'test-user')
        self.assertEqual(form['password'].value(), 'test-pass')

    @add_fixtures(['test_scmtools'])
    def test_save(self):
        """Tetting BZRTool authentication form save"""
        repository = self.create_repository(tool_name='Bazaar')

        form = BZRTool.create_auth_form(
            repository=repository,
            data={
                'username': 'test-user',
                'password': 'test-pass',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.username, 'test-user')
        self.assertEqual(repository.password, 'test-pass')


class BZRRepositoryFormTests(TestCase):
    """Unit tests for BZRTool's repository form."""

    def test_fields(self):
        """Testing BZRTool repository form fields"""
        form = BZRTool.create_repository_form()

        self.assertEqual(list(form.fields), ['path', 'mirror_path'])
        self.assertEqual(form['path'].help_text,
                         'The path to the repository. This will generally be '
                         'the URL you would use to check out the repository.')
        self.assertEqual(form['path'].label, 'Path')
        self.assertEqual(form['mirror_path'].help_text, '')
        self.assertEqual(form['mirror_path'].label, 'Mirror Path')

    @add_fixtures(['test_scmtools'])
    def test_load(self):
        """Tetting BZRTool repository form load"""
        repository = self.create_repository(
            tool_name='Bazaar',
            path='bzr+ssh://bzr.example.com/repo',
            mirror_path='sftp://bzr.example.com/repo')

        form = BZRTool.create_repository_form(repository=repository)
        form.load()

        self.assertEqual(form['path'].value(),
                         'bzr+ssh://bzr.example.com/repo')
        self.assertEqual(form['mirror_path'].value(),
                         'sftp://bzr.example.com/repo')

    @add_fixtures(['test_scmtools'])
    def test_save(self):
        """Tetting BZRTool repository form save"""
        repository = self.create_repository(tool_name='Bazaar')

        form = BZRTool.create_repository_form(
            repository=repository,
            data={
                'path': 'bzr+ssh://bzr.example.com/repo',
                'mirror_path': 'sftp://bzr.example.com/repo',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path, 'bzr+ssh://bzr.example.com/repo')
        self.assertEqual(repository.mirror_path, 'sftp://bzr.example.com/repo')



"""A comment made on a diff."""

from __future__ import annotations

from typing import Any, Optional
from urllib.parse import urlencode

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.translation import gettext_lazy as _

from reviewboard.diffviewer.models import FileDiff
from reviewboard.reviews.models.base_comment import BaseComment
from reviewboard.site.urlresolvers import local_site_reverse


class Comment(BaseComment):
    """A comment made on a diff.

    A comment can belong to a single filediff or to an interdiff between
    two filediffs. It can also have multiple replies.
    """

    _BASE_FILEDIFF_ID_KEY = '__base_filediff_id'

    anchor_prefix = 'comment'
    comment_type = 'diff'
    filediff = models.ForeignKey(FileDiff,
                                 on_delete=models.CASCADE,
                                 verbose_name=_('file diff'),
                                 related_name='comments')
    interfilediff = models.ForeignKey(FileDiff,
                                      on_delete=models.CASCADE,
                                      verbose_name=_('interdiff file'),
                                      blank=True, null=True,
                                      related_name='interdiff_comments')

    # A null line number applies to an entire diff.  Non-null line numbers are
    # the line within the entire file, starting at 1.
    first_line = models.PositiveIntegerField(_('first line'), blank=True,
                                             null=True)
    num_lines = models.PositiveIntegerField(_('number of lines'), blank=True,
                                            null=True)

    last_line = property(lambda self: self.first_line + self.num_lines - 1)

    @property
    def base_filediff_id(self) -> Optional[int]:
        """The base FileDiff ID for the cumulative diff this comment is on.

        Type:
            int
        """
        if self.extra_data:
            return self.extra_data.get(self._BASE_FILEDIFF_ID_KEY)

        return None

    @base_filediff_id.setter
    def base_filediff_id(self, filediff_id: int):
        if self.extra_data is None:
            self.extra_data = {}

        self.extra_data[self._BASE_FILEDIFF_ID_KEY] = filediff_id

    @property
    def base_filediff(self) -> Optional[FileDiff]:
        """The base filediff, if this comment is made on a commit range.

        Type:
            reviewboard.diffviewer.models.FileDiff
        """
        base_filediff_id = self.base_filediff_id

        if base_filediff_id:
            try:
                return FileDiff.objects.get(pk=base_filediff_id)
            except ObjectDoesNotExist:
                pass

        return None

    def get_absolute_url(self) -> str:
        """Return the URL for the given comment.

        Returns:
            str:
            The URL to view the part of the file where the comment was added.
        """
        review_request = self.get_review_request()

        query: dict[str, Any] = {}

        base_filediff = self.base_filediff

        if base_filediff:
            query['base-commit-id'] = base_filediff.commit_id

        if self.filediff.commit_id:
            query['tip-commit-id'] = self.filediff.commit_id

        if self.interfilediff:
            url = local_site_reverse(
                'view-interdiff',
                local_site=review_request.local_site,
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': self.filediff.diffset.revision,
                    'interdiff_revision':
                        self.interfilediff.diffset.revision,
                })
        else:
            url = local_site_reverse(
                'view-diff-revision',
                local_site=review_request.local_site,
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': self.filediff.diffset.revision,
                })

        anchor = '#file%sline%s' % (self.filediff.id, self.first_line)

        return '%s?%s%s' % (url, urlencode(query), anchor)

    def diff_is_public(self) -> bool:
        """Return whether the diff(s) being commented on are public.

        Returns:
            bool:
            True if the diff (and interdiff, if applicable) is public.
        """
        return (self.filediff.diffset.history is not None and
                (self.interfilediff is None or
                 self.interfilediff.diffset.history is not None))

    class Meta(BaseComment.Meta):
        db_table = 'reviews_comment'
        verbose_name = _('Diff Comment')
        verbose_name_plural = _('Diff Comments')



import logging
import os
from tempfile import mkdtemp
from unittest import SkipTest

from paramiko.ssh_exception import NoValidConnectionsError

from reviewboard.scmtools.core import HEAD
from reviewboard.scmtools.errors import SCMError, AuthenticationError
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite
from reviewboard.ssh.client import SSHClient
from reviewboard.ssh.tests import SSHTestCase


logger = logging.getLogger(__name__)


class SCMTestCase(SSHTestCase):
    """Base class for test suites for SCMTools."""

    ssh_client = None

    #: Executables that must be available system-wide for SSH tests.
    #:
    #: SSH tests often require running a command over SSH. These commands
    #: may be available in the local virtualenv where development is taking
    #: place, but may not be available system-wide.
    #:
    #: If this is specified, and the command is not available in the system
    #: path when connecting, the test will be skipped.
    #:
    #: Version Added:
    #:     5.0
    #:
    #: Type:
    #:     list of unicode
    ssh_required_system_exes = None

    _can_test_ssh = None
    _ssh_system_exe_status = {}

    def setUp(self):
        super(SCMTestCase, self).setUp()
        self.tool = None

    def _check_can_test_ssh(self):
        """Check whether SSH-based tests can be run.

        This will check if the user's SSH keys are authorized by the local
        machine for authentication, and whether any system-wide tools are
        available.

        If SSH-based tests cannot be run, the current test will be flagged
        as skipped.
        """
        # These tests are global across all unit tests using this class.
        if SCMTestCase._can_test_ssh is None:
            SCMTestCase.ssh_client = SSHClient()
            key = self.ssh_client.get_user_key()
            SCMTestCase._can_test_ssh = (
                key is not None and
                self.ssh_client.is_key_authorized(key))

        if not SCMTestCase._can_test_ssh:
            raise SkipTest(
                "Cannot perform SSH access tests. The local user's SSH "
                "public key must be in the %s file and SSH must be enabled."
                % os.path.join(self.ssh_client.storage.get_ssh_dir(),
                               'authorized_keys'))

        # These tests are local to all unit tests using the same executable.
        system_exes = self.ssh_required_system_exes

        if system_exes:
            user_key = SCMTestCase.ssh_client.get_user_key()

            exes_to_check = (
                set(system_exes) -
                set(SCMTestCase._ssh_system_exe_status.keys()))

            for system_exe in exes_to_check:
                # For safety, we'll do one connection per check, to avoid
                # one check impacting another.
                client = SSHClient()
                client.connect('localhost',
                               pkey=user_key)

                try:
                    stdout, stderr = client.exec_command('which %s'
                                                         % system_exe)[1:]

                    # It's important to read all stdout/stderr data before
                    # waiting for status.
                    stdout.read()
                    stderr.read()
                    code = stdout.channel.recv_exit_status()

                    status = (code == 0)
                except Exception as e:
                    logger.error('Unexpected error running `which %s` on '
                                 'localhost for SSH test: %s',
                                 system_exe, e)
                    status = False
                finally:
                    client.close()

                SCMTestCase._ssh_system_exe_status[system_exe] = status

            missing_exes = ', '.join(
                '"%s"' % _system_exe
                for _system_exe in system_exes
                if not SCMTestCase._ssh_system_exe_status[_system_exe]
            )

            if missing_exes:
                raise SkipTest(
                    'Cannot perform SSH access tests. %s must be '
                    'available in the system path when executing '
                    'commands locally over SSH. You may need to install the '
                    'tool or make sure that the correct directory is in '
                    '~/.zshenv, ~/.profile, or another suitable file used '
                    'in non-interactive sessions.'
                    % missing_exes)

    def _test_ssh(self, repo_path, filename=None):
        """Helper for testing an SSH connection to a local repository.

        This will attempt to SSH into the local machine and connect to the
        given repository, checking it for validity and optionally fetching
        a file.

        If this is unable to connect to the local machine, the test will be
        flagged as skipped.

        Args:
            repo_path (unicode):
                The repository path to check.

            filename (unicode, optional):
                The optional file in the repository to fetch.
        """
        self._check_can_test_ssh()

        repo = Repository(
            name='SSH Test',
            path=repo_path,
            tool=self.repository.tool,
            scmtool_id=self.repository.scmtool_id)
        tool = repo.get_scmtool()

        try:
            tool.check_repository(repo_path)
        except NoValidConnectionsError:
            # This box likely isn't set up for this test.
            SCMTestCase._can_test_ssh = False

            raise SkipTest(
                'Cannot perform SSH access tests. No local SSH service is '
                'running.')

        if filename:
            self.assertIsNotNone(tool.get_file(filename, HEAD))

    def _test_ssh_with_site(self, repo_path, filename=None):
        """Helper for testing an SSH connection and using a Local Site.

        This will attempt to SSH into the local machine and connect to the
        given repository, using an SSH key and repository based on a Local
        Site. It will check the repository for validity and optionally fetch
        a file.

        If this is unable to connect to the local machine, the test will be
        flagged as skipped.

        Args:
            repo_path (unicode):
                The repository path to check.

            filename (unicode, optional):
                The optional file in the repository to fetch.
        """
        self._check_can_test_ssh()

        # Get the user's .ssh key, for use in the tests
        user_key = self.ssh_client.get_user_key()
        self.assertIsNotNone(user_key)

        # Switch to a new SSH directory.
        self.tempdir = mkdtemp(prefix='rb-tests-home-')
        sshdir = os.path.join(self.tempdir, '.ssh')
        self._set_home(self.tempdir)

        self.assertEqual(sshdir, self.ssh_client.storage.get_ssh_dir())
        self.assertFalse(os.path.exists(os.path.join(sshdir, 'id_rsa')))
        self.assertFalse(os.path.exists(os.path.join(sshdir, 'id_dsa')))
        self.assertIsNone(self.ssh_client.get_user_key())

        tool_class = self.repository.tool

        # Make sure we aren't using the old SSH key. We want auth errors.
        repo = self.create_repository(name='SSH Test',
                                      path=repo_path,
                                      tool_name=tool_class.name)
        tool = repo.get_scmtool()
        self.assertRaises(AuthenticationError,
                          lambda: tool.check_repository(repo_path))

        if filename:
            self.assertRaises(SCMError,
                              lambda: tool.get_file(filename, HEAD))

        for local_site_name in ('site-1',):
            local_site = LocalSite(name=local_site_name)
            local_site.save()

            repo = self.create_repository(
                name='SSH Test',
                path=repo_path,
                tool_name=tool_class.name,
                local_site=local_site)
            tool = repo.get_scmtool()

            ssh_client = SSHClient(namespace=local_site_name)
            self.assertEqual(ssh_client.storage.get_ssh_dir(),
                             os.path.join(sshdir, local_site_name))
            ssh_client.import_user_key(user_key)
            self.assertEqual(ssh_client.get_user_key(), user_key)

            # Make sure we can verify the repository and access files.
            tool.check_repository(repo_path, local_site_name=local_site_name)

            if filename:
                self.assertIsNotNone(tool.get_file(filename, HEAD))



"""A form that handles uploading of new files."""

from __future__ import annotations

import os
from typing import Optional, TYPE_CHECKING, cast
from uuid import uuid4

from django import forms
from django.core.files import File
from django.utils.translation import gettext as _
from djblets.db.fields.json_field import JSONFormField
from djblets.siteconfig.models import SiteConfiguration
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.attachments.errors import FileTooBigError
from reviewboard.attachments.mimetypes import get_uploaded_file_mimetype
from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.diffviewer.diffutils import get_sha256
from reviewboard.reviews.models import ReviewRequestDraft

if TYPE_CHECKING:
    from reviewboard.diffviewer.models import FileDiff
    from reviewboard.reviews.models import ReviewRequest


class UploadFileForm(forms.Form):
    """A form that handles uploading of new files."""

    #: The caption for the file.
    caption = forms.CharField(required=False)

    #: The file itself.
    path = forms.FileField(required=True)

    #: An optional file attachment history.
    #:
    #: This is used when creating a new revision for an existing file
    #: attachment. If this is not specified, a new history will be created.
    attachment_history = forms.ModelChoiceField(
        queryset=FileAttachmentHistory.objects.all(),
        required=False)

    #: Extra data as part of the file attachment.
    extra_data = JSONFormField(required=False)

    def __init__(self, review_request, *args, **kwargs):
        """Initialize the form.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to attach the file to.

            args (tuple):
                Extra positional arguments for the form.

            **kwargs (dict):
                Extra keyword arguments for the form.
        """
        super(UploadFileForm, self).__init__(*args, **kwargs)

        self.review_request = review_request

    def clean_attachment_history(self):
        """Validate that the specified file attachment history exists.

        Returns:
            reviewboard.attachments.models.FileAttachmentHistory:
            The history model.
        """
        history = self.cleaned_data['attachment_history']

        if (history is not None and
            not self.review_request.file_attachment_histories.filter(
                pk=history.pk).exists()):
            raise forms.ValidationError(
                'The FileAttachmentHistory provided is not part of this '
                'review request.')

        return history

    @deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
    def create(
        self,
        *,
        filediff: Optional[FileDiff] = None,
        from_modified: bool = True,
    ) -> FileAttachment:
        """Create a FileAttachment based on this form.

        Version Changed:
            7.0:
            * Added the ``from_modified`` argument.
            * Made arguments keyword-only.

        Args:
            filediff (reviewboard.diffviewer.models.filediff.FileDiff,
                      optional):
                The optional diff to attach this file to (for use when this
                file represents a binary file within the diff).

            from_modified (bool, optional):
                If creating an attachment for a FileDiff and this is ``True``,
                create the attachment for the modified version of the file.
                Otherwise, create it for the source version of the file.

                Version Added:
                    7.0

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment model.

        Raises:
            reviewboard.attachments.errors.FileTooBigError:
                The file is too big for configured limits.
        """
        file_obj = cast(File, self.files['path'])
        review_request: ReviewRequest = self.review_request

        if filediff is not None:
            siteconfig = SiteConfiguration.objects.get_current()
            max_attachment_size = siteconfig.get('diffviewer_max_binary_size')

            if file_obj.size > max_attachment_size:
                raise FileTooBigError(
                    _('The given file is too large.'),
                    max_attachment_size=max_attachment_size,
                )

        caption = self.cleaned_data['caption'] or file_obj.name
        extra_data = self.cleaned_data['extra_data']

        if not extra_data:
            extra_data = {}

        extra_data['sha256_checksum'] = get_sha256(file_obj)
        extra_data[FileAttachment.DEFINED_LOCAL_SITE_KEY] = True

        mimetype = get_uploaded_file_mimetype(file_obj)
        filename = get_unique_filename(file_obj.name)

        attachment_kwargs = {
            'caption': '',
            'draft_caption': caption,
            'extra_data': extra_data,
            'local_site': review_request.local_site,
            'orig_filename': os.path.basename(file_obj.name),
            'mimetype': mimetype,
        }

        if filediff:
            file_attachment = FileAttachment.objects.create_from_filediff(
                filediff,
                from_modified=from_modified,
                **attachment_kwargs)
        else:
            attachment_history = self.cleaned_data['attachment_history']

            if attachment_history is None:
                # This is a new file: create a new FileAttachmentHistory for it
                attachment_history = FileAttachmentHistory()

                attachment_history.display_position = \
                    FileAttachmentHistory.compute_next_display_position(
                        review_request)
                attachment_history.save()
                review_request.file_attachment_histories.add(
                    attachment_history)

                attachment_kwargs['attachment_history'] = attachment_history
                attachment_kwargs['attachment_revision'] = 1
            else:
                try:
                    latest = attachment_history.file_attachments.latest()
                except FileAttachment.DoesNotExist:
                    latest = None

                if latest is None:
                    # This should theoretically never happen, but who knows.
                    attachment_revision = 1
                elif latest.review_request.exists():
                    # This is a new update in the draft.
                    attachment_revision = latest.attachment_revision + 1
                else:
                    # The most recent revision is part of the same draft.
                    # Delete it and replace with the newly uploaded file.
                    attachment_revision = latest.attachment_revision
                    latest.delete()

                attachment_kwargs['attachment_history'] = attachment_history
                attachment_kwargs['attachment_revision'] = attachment_revision

            file_attachment = FileAttachment(**attachment_kwargs)

        file_attachment._review_request = review_request
        file_attachment.file.save(filename, file_obj, save=True)

        if not filediff:
            draft = ReviewRequestDraft.create(review_request)
            draft.file_attachments.add(file_attachment)
            draft.save()

        return file_attachment


class UploadUserFileForm(forms.Form):
    """A form that handles uploading of user files."""

    #: The caption for the file.
    caption = forms.CharField(required=False)

    #: The file itself.
    path = forms.FileField(required=False)

    #: Extra data as part of the file attachment.
    extra_data = JSONFormField(required=False)

    def create(self, user, local_site=None):
        """Create a FileAttachment based on this form.

        Args:
            user (django.contrib.auth.models.User):
                The user who owns this file attachment.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional local site.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment model.
        """
        file_obj = self.files.get('path')

        extra_data = self.cleaned_data['extra_data']

        if not extra_data:
            extra_data = {}

        extra_data[FileAttachment.DEFINED_LOCAL_SITE_KEY] = True

        attachment_kwargs = {
            'extra_data': extra_data,
            'local_site': local_site,
            'uuid': uuid4(),
            'user': user,
        }

        if file_obj:
            mimetype = get_uploaded_file_mimetype(file_obj)
            filename = get_unique_filename(file_obj.name)

            attachment_kwargs.update({
                'caption': self.cleaned_data['caption'] or file_obj.name,
                'orig_filename': os.path.basename(file_obj.name),
                'mimetype': mimetype,
            })

            file_attachment = FileAttachment(**attachment_kwargs)
            file_attachment.file.save(filename, file_obj, save=True)
        else:
            attachment_kwargs['caption'] = self.cleaned_data['caption'] or ''

            file_attachment = FileAttachment.objects.create(
                **attachment_kwargs)

        return file_attachment

    def update(self, file_attachment):
        """Update an existing file attachment.

        Args:
            file_attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment to update.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The updated file attachment.
        """
        caption = self.cleaned_data['caption']
        file_obj = self.files.get('path')
        extra_data = self.cleaned_data['extra_data']

        if caption:
            file_attachment.caption = caption

        if file_obj:
            file_attachment.mimetype = get_uploaded_file_mimetype(file_obj)
            file_attachment.orig_filename = os.path.basename(file_obj.name)
            file_attachment.file.save(get_unique_filename(file_obj.name),
                                      file_obj, save=True)

        if extra_data:
            file_attachment.extra_data.update(extra_data)

        file_attachment.save()

        return file_attachment


def get_unique_filename(filename):
    """Return a unique filename.

    Create a unique filename by concatenating a UUID with the given filename.

    Args:
        filename (unicode):
            The original filename.

    Returns:
        unicode:
        A new filename which is more unique.
    """
    return '%s__%s' % (uuid4(), filename)



"""Added FileAttachment.uuid.

Version Added:
    2.5
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('FileAttachment', 'uuid', models.CharField, max_length=255,
             initial=''),
]



"""Upgrade oauth2_provider from 0.9 to 1.6.3.

This is roughly equivalent to the following migrations in oauth2_provider
1.6.3:

* ``0001_initial``
* ``0002_auto_20190406_1805``
* ``0003_auto_20201211_1314``
* ``0004_auto_20200902_2022``
* ``0005_auto_20211222_2352``

These migrations are recorded in
:py:func:`~reviewboard.upgrade.post_upgrade_reset_oauth2_provider`.

Version Added:
    5.0
"""

from django.db import models
from django.utils import timezone
from django_evolution.mutations import AddField, ChangeField, ChangeMeta


MUTATIONS = [
    # Changes to the oauth2_provider_application table.
    AddField('Application', 'created', models.DateTimeField,
             initial=timezone.now),
    AddField('Application', 'updated', models.DateTimeField,
             initial=timezone.now),
    AddField('Application', 'algorithm', models.CharField,
             max_length=5, initial=''),
    ChangeField('Application', 'user', initial=None, null=True),
    ChangeField('Application', 'id', field_type=models.BigAutoField,
                primary_key=True),

    # Changes to the oauth2_provider_grant table.
    AddField('Grant', 'created', models.DateTimeField, initial=timezone.now),
    AddField('Grant', 'updated', models.DateTimeField, initial=timezone.now),
    AddField('Grant', 'code_challenge', models.CharField,
             max_length=128, initial=''),
    AddField('Grant', 'code_challenge_method', models.CharField,
             max_length=10, initial=''),
    AddField('Grant', 'nonce', models.CharField, max_length=255, initial=''),
    AddField('Grant', 'claims', models.TextField, initial=''),
    ChangeField('Grant', 'code', max_length=255, unique=True, db_index=False),
    ChangeField('Grant', 'redirect_uri', field_type=models.TextField),
    ChangeField('Grant', 'id', field_type=models.BigAutoField,
                primary_key=True),

    # Changes to the oauth2_provider_accesstoken table.
    AddField('AccessToken', 'source_refresh_token', models.OneToOneField,
             null=True, unique=True,
             related_model='oauth2_provider.RefreshToken'),
    AddField('AccessToken', 'id_token', models.OneToOneField,
             null=True, unique=True, related_model='oauth2_provider.IDToken'),
    AddField('AccessToken', 'created', models.DateTimeField,
             initial=timezone.now),
    AddField('AccessToken', 'updated', models.DateTimeField,
             initial=timezone.now),
    ChangeField('AccessToken', 'application', null=True),
    ChangeField('AccessToken', 'token', max_length=255, unique=True,
                db_index=False),
    ChangeField('AccessToken', 'id', field_type=models.BigAutoField,
                primary_key=True),

    # Changes to the oauth2_provider_refreshtoken table.
    AddField('RefreshToken', 'created', models.DateTimeField,
             initial=timezone.now),
    AddField('RefreshToken', 'updated', models.DateTimeField,
             initial=timezone.now),
    AddField('RefreshToken', 'revoked', models.DateTimeField, null=True),
    ChangeField('RefreshToken', 'access_token', null=True),
    ChangeField('RefreshToken', 'token', max_length=255, db_index=False),
    ChangeField('RefreshToken', 'id', field_type=models.BigAutoField,
                primary_key=True),
    ChangeMeta('RefreshToken', 'unique_together', [('token', 'revoked')]),
]



"""Template tags for extensions."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import template
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from reviewboard.extensions.hooks import (CommentDetailDisplayHook,
                                          NavigationBarHook)
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.template import Context
    from django.utils.safestring import SafeString

    from reviewboard.reviews.models import BaseComment


logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(takes_context=True)
def navigation_bar_hooks(
    context: Context,
    *,
    sidebar: bool = False,
) -> SafeString:
    """Display all registered navigation bar entries.

    Version Changed:
        7.0.3:
        * Added the ``sidebar`` argument.

    Args:
        context (django.template.Context):
            The current template render context.

        sidebar (bool, optional):
            Whether the navigation bar entries are being rendered in sidebar
            form. If ``False``, the navigation bar entries are being rendered
            in the topbar.

            Version Added:
                7.0.3

    Returns:
        django.utils.safestring.SafeString:
        The rendered navigation bar entries.
    """
    html = []

    if sidebar:
        template_name = 'extensions/navbar_entry_sidebar.html'
    else:
        template_name = 'extensions/navbar_entry.html'

    for hook in NavigationBarHook.hooks:
        try:
            for nav_info in hook.get_entries(context):
                if nav_info:
                    url_name = nav_info.get('url_name', None)

                    if url_name:
                        nav_info['url'] = local_site_reverse(
                            url_name, request=context.get('request'))

                    context.push()
                    context['entry'] = nav_info

                    html.append(render_to_string(
                        template_name=template_name,
                        context=context.flatten()))
                    context.pop()
        except Exception as e:
            extension = hook.extension
            logger.exception(
                'Error when running NavigationBarHook.get_entries function in '
                'extension: "%s": %s',
                extension.id, e)

    return mark_safe(''.join(html))


@register.simple_tag(takes_context=True)
def comment_detail_display_hook(
    context: Context,
    comment: BaseComment,
    render_mode: str,
) -> SafeString:
    """Display all additional detail from CommentDetailDisplayHooks.

    Args:
        context (django.template.Context):
            The current template render context.

        comment (reviewboard.reviews.models.BaseComment):
            The comment being rendered.

        render_mode (str):
            The current render mode. This will be one of "review",
            "text-email", or "html-email", depending on where the comment is
            being rendered.

    Returns:
        django.utils.safestring.SafeString:
        The rendered navigation bar entries.
    """
    assert render_mode in ('review', 'text-email', 'html-email')

    html = []

    for hook in CommentDetailDisplayHook.hooks:
        try:
            if render_mode == 'review':
                html.append(hook.render_review_comment_detail(comment))
            elif render_mode in ('text-email', 'html-email'):
                html.append(hook.render_email_comment_detail(
                    comment, render_mode == 'html-email'))
        except Exception as e:
            extension = hook.extension
            logger.exception(
                'Error when running CommentDetailDisplayHook with render mode '
                '"%s" in extension: %s: %s',
                render_mode, extension.id, e)

    return mark_safe(''.join(html))



"""Unit tests for reviewboard.admin.form_widgets.RelatedGroupWidget."""

from django import forms

from reviewboard.admin.form_widgets import RelatedGroupWidget
from reviewboard.reviews.models import Group
from reviewboard.testing.testcase import TestCase


class TestForm(forms.Form):
    """A Test Form with a field that contains a RelatedGroupWidget."""
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=Group.objects.filter(visible=True).order_by('name'),
        label=('Default groups'),
        required=False,
        widget=RelatedGroupWidget())


class LocalSiteTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedGroupWidget.

    The RelatedGroupWidget is defined to have a local_site_name.
    """
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=Group.objects.filter(visible=True).order_by('name'),
        label=('Default groups'),
        required=False,
        widget=RelatedGroupWidget(local_site_name='supertest'))


class SingleValueTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedGroupWidget.

    RelatedGroupWidget is defined as setting multivalued to False.
    """
    my_select_field = forms.ModelMultipleChoiceField(
        queryset=Group.objects.filter(visible=True).order_by('name'),
        label=('Default groups'),
        required=False,
        widget=RelatedGroupWidget(multivalued=False))


class RelatedGroupWidgetTests(TestCase):
    """Unit tests for RelatedGroupWidget."""

    def test_render_empty(self):
        """Testing RelatedGroupWidget.render with no initial data"""
        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default groups',
            [],
            {'id': 'groups'})

        self.assertHTMLEqual(
            """<input id="groups" name="Default groups" type="hidden" />

            <script>
            $(function() {
                var view = new RB.RelatedGroupSelectorView({
                    $input: $('#groups'),
                    "initialOptions": [],

                    "inviteOnly": false,
                    "multivalued": true,
                    "showInvisible": true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_data(self):
        """Testing RelatedGroupWidget.render with initial data"""
        test_group_1 = self.create_review_group(name='group1')
        test_group_2 = self.create_review_group(name='group2')
        test_group_3 = self.create_review_group(name='group3')

        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default groups',
            [test_group_1.pk, test_group_2.pk, test_group_3.pk],
            {'id': 'groups'})

        self.assertHTMLEqual(
            """<input id="groups" name="Default groups" type="hidden"
                      value="1,2,3" />

            <script>
            $(function() {
                var view = new RB.RelatedGroupSelectorView({
                    $input: $('#groups'),
                    "initialOptions":
                    [{"display_name": "", "id": 1, "name": "group1"},
                    {"display_name": "", "id": 2, "name": "group2"},
                    {"display_name": "", "id": 3, "name": "group3"}],

                    "inviteOnly": false,
                    "multivalued": true,
                    "showInvisible": true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_local_site(self):
        """Testing RelatedGroupWidget.render with a local site defined"""
        my_form = LocalSiteTestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default groups',
            [],
            {'id': 'groups'})

        self.assertIn('"localSitePrefix": "s/supertest/",', html)

    def test_value_from_datadict(self):
        """Testing RelatedGroupWidget.value_from_datadict"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'groups': ['1', '2']},
                {},
                'groups'))

        self.assertEqual(value, ['1', '2'])

    def test_value_from_datadict_single_value(self):
        """Testing RelatedGroupWidget.value_from_datadict with single value"""
        my_form = SingleValueTestForm()
        value = (
            my_form.fields['my_select_field']
            .widget
            .value_from_datadict(
                {'groups': ['1']},
                {},
                'groups'))

        self.assertEqual(value, ['1'])

    def test_value_from_datadict_with_no_data(self):
        """Testing RelatedGroupWidget.value_from_datadict with no data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'groups': []},
                {},
                'groups'))

        self.assertEqual(value, [])

    def test_value_from_datadict_with_missing_data(self):
        """Testing RelatedGroupWidget.value_from_datadict with missing data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {},
                {},
                'groups'))
        self.assertIsNone(value)



"""PySVN client backend for Subversion."""

from __future__ import annotations

import logging
import os
from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime
from shutil import rmtree
from tempfile import mkdtemp
from typing import (Any, Dict, Iterator, Optional, Sequence, TYPE_CHECKING,
                    Tuple)

try:
    import pysvn
    from pysvn import ClientError, Revision, opt_revision_kind
    has_svn_backend = True
except ImportError:
    # This try-except block is here for the sole purpose of avoiding
    # exceptions with nose if pysvn isn't installed when someone runs
    # the testsuite.
    has_svn_backend = False

from django.utils.encoding import force_bytes, force_str
from django.utils.translation import gettext as _

from reviewboard.scmtools.core import HEAD, PRE_CREATION
from reviewboard.scmtools.errors import (AuthenticationError,
                                         FileNotFoundError,
                                         SCMError)
from reviewboard.scmtools.svn import base
from reviewboard.scmtools.svn.utils import (collapse_svn_keywords,
                                            has_expanded_svn_keywords)

if TYPE_CHECKING:
    from reviewboard.scmtools.core import RevisionID
    from reviewboard.scmtools.svn.base import (AcceptCertificateFunc,
                                               SVNDirEntry,
                                               SVNLogEntry,
                                               SVNRepositoryInfoDict,
                                               SSLServerTrustPromptFunc)


logger = logging.getLogger(__name__)


class Client(base.Client):
    """PySVN-based Subversion client backend.

    This provides Subversion compatibility through the well-maintained
    `PySVN <https://pysvn.sourceforge.io/>`_ library.
    """

    required_module = 'pysvn'

    ######################
    # Instance variables #
    ######################

    #: The PySVN client used for communication.
    #:
    #: Type:
    #:     pysvn.Client
    client: pysvn.Client

    def __init__(
        self,
        config_dir: str,
        repopath: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """Initialize the client.

        Args:
            config_dir (str):
                The path to the Subversion configuration directory.

            repopath (str):
                The path to the repository.

            username (str, optional):
                The username for the repository.

            password (str, optional):
                The password for the repository.
        """
        super().__init__(config_dir=config_dir,
                         repopath=repopath,
                         username=username,
                         password=password)

        client = pysvn.Client(config_dir)

        if username:
            client.set_default_username(username)

        if password:
            client.set_default_password(password)

        self.client = client

    def normalize_error(
        self,
        e: Exception,
        *,
        default_msg: Optional[str] = None,
    ) -> Exception:
        """Normalize an exception from the client.

        This will process the exception information and return an exception
        suitable for forwarding to the backend, as documented below.

        Args:
            e (Exception):
                The exception to normalize.

            default_msg (str, optional):
                An optional default message to use for a default exception.

        Returns:
            Exception:
            The normalized exception.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                An authentication error communicating with the repository.

            reviewboard.scmtools.errors.SCMError:
                A general error communicating with the repository.
        """
        if isinstance(e, ClientError):
            msg = str(e)

            if 'callback_get_login required' in msg:
                return AuthenticationError(msg=_(
                    'Authentication failed when talking to the Subversion '
                    'repository.'
                ))
            elif 'callback_ssl_server_trust_prompt required' in msg:
                return SCMError(
                    _('HTTPS certificate not accepted. Please ensure that '
                      'the proper certificate exists in %s '
                      'for the user that Review Board is running as.')
                    % os.path.join(self.config_dir, 'auth'))

        return SCMError(default_msg or str(e))

    def set_ssl_server_trust_prompt(
        self,
        cb: SSLServerTrustPromptFunc,
    ) -> None:
        """Set a function for verifying a SSL certificate.

        Args:
            cb (callable):
                The function to call for verifying SSL certificates.
        """
        self.client.callback_ssl_server_trust_prompt = cb

    @contextmanager
    def _do_on_path(
        self,
        path: str,
        revision: RevisionID = HEAD,
    ) -> Iterator[Tuple[str, str]]:
        """Perform an operation on a given path.

        This will normalize the provided path and revision and then call the
        given function with those normalized values. The error will be
        normalized and any errors converted into useful exceptions.

        Args:
            cb (callable):
                The function to call.

            path (str):
                The repository path to provide to the callback.

            revision (str):
                The revision to provide to the callback.

        Returns:
            object:
            The value from the callback.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                There was an error authenticating with the repository.

            reviewboard.scmtools.errors.FileNotFoundError:
                The path was empty or could not be found in the repository.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in an invalid format.

            reviewboard.scmtools.errors.SCMError:
                There was an error communicating with the repository.
        """
        if not path:
            raise FileNotFoundError(path, revision)

        with self.communicate():
            try:
                yield (self.normalize_path(path),
                       self._normalize_revision(revision))
            except ClientError as e:
                msg = force_str(e)

                if 'File not found' in msg or 'path not found' in msg:
                    raise FileNotFoundError(path, revision, detail=msg)

                raise

    def get_file(
        self,
        path: str,
        revision: RevisionID = HEAD,
    ) -> bytes:
        """Return the contents of a file from the repository.

        This attempts to return the raw binary contents of a file from the
        repository, given a file path and revision.

        If the contents have expanded keywords, they'll be collapsed.

        Args:
            path (str):
                The path to the file in the repository.

            revision (reviewboard.scmtools.core.Revision, optional):
                The revision to fetch. Subclasses should default this to
                :py:data:`HEAD`.

            *args (tuple, unused):
                Additional unused positional arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Returns:
            bytes:
            The returned file contents.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found in the repository.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in an invalid format.

            reviewboard.scmtools.errors.SCMError:
                An unexpected error was encountered with the repository.
        """
        client = self.client

        with self._do_on_path(path, revision) as (path, revision):
            data = client.cat(path, revision)

            if has_expanded_svn_keywords(data):
                # Find out if this file has any keyword expansion set. If it
                # does, collapse these keywords. This is because SVN will
                # return the file expanded to us, which would break patching.
                keywords = client.propget('svn:keywords', path, revision,
                                          recurse=True)

                if path in keywords:
                    data = collapse_svn_keywords(data,
                                                 force_bytes(keywords[path]))

        return data

    def get_keywords(
        self,
        path: str,
        revision: RevisionID = HEAD,
    ) -> str:
        """Return a space-separated list of SVN keywords for a given path.

        Args:
            path (str):
                The path to the file or directory.

            revision (str):
                The revision containing the keywords.

        Returns:
            str:
            The resulting keywords.
        """
        with self._do_on_path(path, revision) as (path, revision):
            return (
                self.client.propget('svn:keywords', path, revision,
                                    recurse=True)
                .get(path)
            )

    def _normalize_revision(
        self,
        revision: RevisionID,
    ) -> Revision:
        """Return a normalized revision for PySVN.

        Args:
            revision (str or reviewboard.scmtools.core.Revision):
                The revision to normalize.

        Returns:
            pysvn.Revision:
            The normalized revision.
        """
        if revision == HEAD:
            r = Revision(opt_revision_kind.head)
        elif revision == PRE_CREATION:
            raise FileNotFoundError('', revision)
        else:
            r = Revision(opt_revision_kind.number, str(revision))

        return r

    @property
    def repository_info(self) -> SVNRepositoryInfoDict:
        """Metadata about the repository.

        See :py:class:`reviewboard.scmtools.svn.base.SVNRepositoryInfoDict`
        for contents.

        Type:
            dict
        """
        with self.communicate():
            info = self.client.info2(self.repopath, recurse=False)

        return {
            'uuid': info[0][1].repos_UUID,
            'root_url': info[0][1].repos_root_URL,
            'url': info[0][1].URL
        }

    def accept_ssl_certificate(
        self,
        path: str,
        on_failure: Optional[AcceptCertificateFunc] = None,
    ) -> Dict[str, Any]:
        """Attempt to accept a SSL certificate.

        If the repository uses SSL, this method is used to determine whether
        the SSL certificate can be automatically accepted.

        If the cert cannot be accepted, the ``on_failure`` callback
        is executed.

        Args:
            path (str):
                The repository path.

            on_failure (callable, optional):
                A function to call if the certificate could not be accepted.

        Returns:
            dict:
            Serialized information on the certificate.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                There was an error accepting the certificate.
        """
        cert: Dict[str, Any] = {}

        def ssl_server_trust_prompt(trust_dict):
            cert.update(trust_dict.copy())

            if on_failure:
                return False, 0, False
            else:
                del cert['failures']
                return True, trust_dict['failures'], True

        self.client.callback_ssl_server_trust_prompt = ssl_server_trust_prompt

        try:
            info = self.client.info2(path, recurse=False)
            logger.debug('SVN: Got repository information for %s: %s',
                         path, info)
        except Exception as e:
            if on_failure:
                on_failure(self, e, path, cert)

        return cert

    def get_log(
        self,
        path: str,
        *,
        start: Optional[str] = None,
        end: Optional[str] = None,
        limit: Optional[int] = None,
        discover_changed_paths: bool = False,
        limit_to_path: bool = False,
    ) -> Sequence[SVNLogEntry]:
        """Return log entries at the specified path.

        The log entries will appear ordered from most recent to least,
        with ``start`` being the most recent commit in the range.

        Args:
            path (str):
                The path to limit log entries to.

            start (str, optional):
                The start revision for the log.

                If not specified, clients must default this to "HEAD".

            end (str, optional):
                The end revision for the log.

                If not specified, clients must default this to "1".

            limit (int, optional):
                The maximum number of entries to return.

            discover_changed_paths (bool, optional):
                Whether to include information on changed paths in the
                results.

            limit_to_path (bool, optional):
                Limits results to ``path`` without factoring in history from
                any branch operations.

        Returns:
            list of SVNLogEntry:
            The list of resulting log entries.
        """
        if start is None:
            start = self.LOG_DEFAULT_START

        if end is None:
            end = self.LOG_DEFAULT_END

        with self.communicate():
            commits = self.client.log(
                self.normalize_path(path),
                limit=limit,
                revision_start=self._normalize_revision(start),
                revision_end=self._normalize_revision(end),
                discover_changed_paths=discover_changed_paths,
                strict_node_history=limit_to_path)

        for commit in commits:
            commit['revision'] = str(commit['revision'].number)

            if 'date' in commit:
                commit['date'] = datetime.utcfromtimestamp(commit['date'])

        return commits

    def list_dir(
        self,
        path: str,
    ) -> OrderedDict[str, SVNDirEntry]:
        """Return the directory contents of the specified path.

        The result will be an ordered dictionary of contents, mapping
        filenames or directory names with directory information.

        Args:
            path (str):
                The directory path.

        Returns:
            collections.OrderedDict:
            A dictionary mapping directory names as strings to
            :py:class:`SVNDirEntry` information.
        """
        result: OrderedDict[str, SVNDirEntry] = OrderedDict()
        norm_path = self.normalize_path(path)

        with self.communicate():
            dirents = self.client.list(norm_path, recurse=False)[1:]

        repo_path_len = len(self.repopath)

        for dirent, unused in dirents:
            name = dirent['path'].split('/')[-1]

            result[name] = {
                'path': dirent['path'][repo_path_len:],
                'created_rev': str(dirent['created_rev'].number),
            }

        return result

    def diff(
        self,
        revision1: str,
        revision2: str,
        path: Optional[str] = None,
    ) -> bytes:
        """Return a diff between two revisions.

        The diff will contain the differences between the two revisions,
        and may optionally be limited to a specific path.

        Args:
            revision1 (str):
                The first revision in the range.

            revision2 (str):
                The second revision in the range.

            path (str, optional):
                An optional path to limit the diff to.

        Returns:
            bytes:
            The resulting diff contents.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                An error was encountered with the repository.

                This may be a more specific subclass.
        """
        if path:
            path = self.normalize_path(path)
        else:
            path = self.repopath

        with self.communicate():
            tmpdir = mkdtemp(prefix='reviewboard-svn.')

            try:
                diff = force_bytes(self.client.diff(
                    tmpdir,
                    path,
                    revision1=self._normalize_revision(revision1),
                    revision2=self._normalize_revision(revision2),
                    header_encoding='UTF-8',
                    diff_options=['-u']))
            except Exception as e:
                logger.exception('Failed to generate diff using pysvn for '
                                 'revisions "%s:%s" for path "%s": %s',
                                 revision1, revision2, path, e)

                raise self.normalize_error(
                    e,
                    default_msg=(
                        _('Unable to get diff revisions %(revision1)s '
                          'through %(revision2)s: %(detail)s')
                        % {
                            'detail': e,
                            'revision1': revision1,
                            'revision2': revision1,
                        }
                    ))
            finally:
                rmtree(tmpdir)

        return diff



"""Licensing-related errors.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typelets.django.json import SerializableDjangoJSONDict


class LicenseActionError(Exception):
    """An error result from an action.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: The payload to return in the result.
    payload: SerializableDjangoJSONDict

    def __init__(
        self,
        message: str,
        *,
        payload: (SerializableDjangoJSONDict | None) = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message to display.

            payload (dict, optional):
                Optional payload data to return in the result.
        """
        super().__init__(message)

        self.payload = payload or {}



"""Signal handlers for file attachments.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import Type

from django.db.models.signals import pre_delete

from reviewboard.attachments.models import FileAttachment


def _on_file_attachment_deleted(
    sender: Type[FileAttachment],
    instance: FileAttachment,
    using: str,
    **kwargs,
) -> None:
    """Handle any extra cleanup when a file attachment is deleted.

    This will ensure that any files associated with the file attachment
    are removed from the filesystem.

    Version Added:
        6.0

    Args:
        sender (type, unused):
            The sender of the signal.

        instance (reviewboard.attachments.models.FileAttachment):
            The file attachment being deleted.

        using (str, unused):
            The database alias being used.

        **kwargs (dict, unused):
            Unused additional keyword arguments.
    """
    handler = instance.mimetype_handler

    if handler:
        handler.delete_associated_files()

    instance.file.delete(save=False)


def connect_signal_handlers() -> None:
    """Connect file attachment related signal handlers.

    Version Added:
        6.0
    """
    pre_delete.connect(_on_file_attachment_deleted,
                       sender=FileAttachment)



"""Administration UI customization for models."""

from django.contrib.admin import ModelAdmin as DjangoModelAdmin


class ModelAdmin(DjangoModelAdmin):
    """Base class for administration UI representations of models.

    This should be used instead of :py:class:`django.contrib.admin.ModelAdmin`
    for any new model registrations. It provides additional options for
    Review Board's administration UI.

    Version Added:
        4.0
    """

    #: The template used for any fieldsets on the model's Change Form.
    #:
    #: If not explicitly overridden, ``admin/includes/fieldset.html`` will
    #: be used.
    fieldset_template_name = None



"""Management command to load data into the database."""

import textwrap
import sys

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """Management command to load data into the database."""

    help = (
        "[No longer supported] Loads data formatted by dumpdb.\n"
        "\n"
        "This functionality has been removed. Please use your database's "
        "native tools instead, or contact support@beanbaginc.com for "
        "alternative solutions."
    )

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            'filename',
            metavar='NAME',
            nargs='*',
            help='The name of the file to load.')

    def handle(self, **options):
        """Handle the command.

        Args:
            **options (dict, unused):
                Options parsed on the command line.

        Raises:
            django.core.management.CommandError:
                There was an error with arguments or disabling the extension.
        """
        self.stderr.write('\n')
        self.stderr.write(textwrap.fill(
            "dumpdb and loaddb are no longer supported. They weren't meant "
            "for production installs, and we weren't able to retain "
            "compatibility with the version of Django now used by "
            "Review Board. We recommend using your database's "
            "native SQL dumping and loading tools instead.\n"))
        self.stderr.write('\n')
        self.stderr.write(textwrap.fill(
            "If you need this functionality, or assistance with "
            "transitioning databases, you can contact us at "
            "support@beanbaginc.com for options."))

        sys.exit(1)



"""Unit tests for reviewboard.extensions.hooks.APIExtraDataAccessHook."""

from djblets.registries.errors import AlreadyRegisteredError, RegistrationError
from kgb import SpyAgency

from reviewboard.extensions.base import Extension
from reviewboard.extensions.hooks import APIExtraDataAccessHook
from reviewboard.extensions.tests.testcases import ExtensionHookTestCaseMixin
from reviewboard.webapi.base import ExtraDataAccessLevel, WebAPIResource
from reviewboard.webapi.tests.base import BaseWebAPITestCase


class GenericTestResource(WebAPIResource):
    name = 'test'
    uri_object_key = 'test_id'
    extra_data = {}
    item_mimetype = 'application/vnd.reviewboard.org.test+json'

    fields = {
        'extra_data': {
            'type': dict,
            'description': 'Extra data as part of the test resource. '
                           'This can be set by the API or extensions.',
        },
    }

    allowed_methods = ('GET', 'PUT')

    def get(self, *args, **kwargs):
        return 200, {
            'test': {
                'extra_data': self.serialize_extra_data_field(self)
            }
        }

    def put(self, request, *args, **kwargs):
        fields = request.POST.dict()
        self.import_extra_data(self, self.extra_data, fields)

        return 200, {
            'test': self.extra_data
        }

    def has_access_permissions(self, request, obj, *args, **kwargs):
        return True


class APIExtraDataAccessHookExtension(Extension):
    """Test extension used for APIExtraDataAccessHook tests."""

    resources = [GenericTestResource()]


class EverythingPrivateHook(APIExtraDataAccessHook):
    """Hook which overrides callable to return all fields as private."""

    def get_extra_data_state(self, key_path):
        self.called = True
        return ExtraDataAccessLevel.ACCESS_STATE_PRIVATE


class InvalidCallableHook(APIExtraDataAccessHook):
    """Hook which implements an invalid callable"""

    get_extra_data_state = 'not a callable'


class APIExtraDataAccessHookTests(ExtensionHookTestCaseMixin, SpyAgency,
                                  BaseWebAPITestCase):
    """Testing APIExtraDataAccessHook."""

    extension_class = APIExtraDataAccessHookExtension
    fixtures = ['test_users']

    def setUp(self):
        super(APIExtraDataAccessHookTests, self).setUp()

        self.resource = self.extension.resources[0]
        self.url = self.resource.get_item_url(test_id=1)
        self.resource.extra_data = {
            'public': 'foo',
            'private': 'secret',
            'readonly': 'bar',
        }

    def test_register(self):
        """Testing APIExtraDataAccessHook registration"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('public',), ExtraDataAccessLevel.ACCESS_STATE_PUBLIC)
            ])

        self.assertNotEqual(list(self.resource.extra_data_access_callbacks),
                            [])

    def test_register_overridden_hook(self):
        """Testing APIExtraDataAccessHook with overridden registration for
        resource
        """
        EverythingPrivateHook(self.extension, self.resource, [])

        self.assertNotEqual(list(self.resource.extra_data_access_callbacks),
                            [])

    def test_overridden_hook_get(self):
        """Testing APIExtraDataAccessHook overrides result HTTP GET"""
        hook = EverythingPrivateHook(self.extension, self.resource, [])

        rsp = self.api_get(self.url,
                           expected_mimetype=self.resource.item_mimetype)

        # Since the hook registers the callback function on initialization,
        # which stores a pointer to the method, we can't use SpyAgency after
        # the hook has already been initialized. Since SpyAgency's spy_on
        # function requires an instance of a class, we also cannot spy on the
        # hook function before initialization. Therefore, as a workaround,
        # we're setting a variable in the function to ensure that it is in
        # fact being called.
        self.assertTrue(hook.called)
        self.assertNotIn('public', rsp['test']['extra_data'])
        self.assertNotIn('readonly', rsp['test']['extra_data'])
        self.assertNotIn('private', rsp['test']['extra_data'])

    def test_overridden_hook_put(self):
        """Testing APIExtraDataAccessHook overrides result for HTTP PUT"""
        hook = EverythingPrivateHook(self.extension, self.resource, [])

        original_value = self.resource.extra_data['readonly']
        modified_extra_fields = {
            'extra_data.public': 'modified',
        }

        rsp = self.api_put(self.url, modified_extra_fields,
                           expected_mimetype=self.resource.item_mimetype)

        # Since the hook registers the callback function on initialization,
        # which stores a pointer to the method, we can't use SpyAgency after
        # the hook has already been initialized. Since SpyAgency's spy_on
        # function requires an instance of a class, we also cannot spy on the
        # hook function before initialization. Therefore, as a workaround,
        # we're setting a variable in the function to ensure that it is in
        # fact being called.
        self.assertTrue(hook.called)
        self.assertEqual(original_value, rsp['test']['readonly'])

    def test_register_invalid_hook(self):
        """Testing APIExtraDataAccessHook registration with invalid
        get_extra_data_state
        """
        with self.assertRaises(RegistrationError):
            InvalidCallableHook(self.extension, self.resource, [])

        self.assertSetEqual(set(),
                            set(self.resource.extra_data_access_callbacks))

    def test_register_hook_already_registered(self):
        """Testing APIExtraDataAccessHook registration with already
        registered callback
        """
        hook = APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('public',), ExtraDataAccessLevel.ACCESS_STATE_PUBLIC)
            ])

        with self.assertRaises(AlreadyRegisteredError):
            hook.resource.extra_data_access_callbacks.register(
                hook.get_extra_data_state)

        self.assertNotEqual(set(),
                            set(self.resource.extra_data_access_callbacks))

    def test_public_state_get(self):
        """Testing APIExtraDataAccessHook public state GET"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('public',), ExtraDataAccessLevel.ACCESS_STATE_PUBLIC)
            ])

        rsp = self.api_get(self.url,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertIn('public', rsp['test']['extra_data'])

    def test_public_state_put(self):
        """Testing APIExtraDataAccessHook public state PUT"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('public',), ExtraDataAccessLevel.ACCESS_STATE_PUBLIC)
            ])

        modified_extra_fields = {
            'extra_data.public': 'modified',
        }

        rsp = self.api_put(self.url, modified_extra_fields,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertEqual(modified_extra_fields['extra_data.public'],
                         rsp['test']['public'])

    def test_readonly_state_get(self):
        """Testing APIExtraDataAccessHook readonly state get"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('readonly',),
                 ExtraDataAccessLevel.ACCESS_STATE_PUBLIC_READONLY)
            ])

        rsp = self.api_get(self.url,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertIn('readonly', rsp['test']['extra_data'])

    def test_readonly_state_put(self):
        """Testing APIExtraDataAccessHook readonly state put"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('readonly',),
                 ExtraDataAccessLevel.ACCESS_STATE_PUBLIC_READONLY)
            ])

        original_value = self.resource.extra_data['readonly']
        modified_extra_fields = {
            'extra_data.readonly': 'modified',
        }

        rsp = self.api_put(self.url, modified_extra_fields,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertEqual(original_value, rsp['test']['readonly'])

    def test_private_state_get(self):
        """Testing APIExtraDataAccessHook private state get"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('private',), ExtraDataAccessLevel.ACCESS_STATE_PRIVATE)
            ])

        rsp = self.api_get(self.url,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertNotIn('private', rsp['test']['extra_data'])

    def test_private_state_put(self):
        """Testing APIExtraDataAccessHook private state put"""
        APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('private',), ExtraDataAccessLevel.ACCESS_STATE_PRIVATE)
            ])

        original_value = self.resource.extra_data['private']
        modified_extra_fields = {
            'extra_data.private': 'modified',
        }

        rsp = self.api_put(self.url, modified_extra_fields,
                           expected_mimetype=self.resource.item_mimetype)

        self.assertEqual(original_value, rsp['test']['private'])

    def test_unregister(self):
        """Testing APIExtraDataAccessHook unregistration"""
        hook = APIExtraDataAccessHook(
            self.extension,
            self.resource,
            [
                (('public',), ExtraDataAccessLevel.ACCESS_STATE_PUBLIC)
            ])

        hook.shutdown()

        self.assertSetEqual(set(),
                            set(self.resource.extra_data_access_callbacks))
